from __future__ import annotations

import os
import shutil
from concurrent.futures import ProcessPoolExecutor
from pathlib import Path
from typing import Callable, Dict, Generic, Iterable, Optional, Tuple, Union

from tqdm import tqdm

from .types import Stringable, T
from .util import is_empty, safe_move


def build_parent(item: T, *ordering: Callable[[T], Stringable]) -> Path:
    parent = Path()

    for getter in ordering:
        parent /= str(getter(item))

    return parent


class Tree(Generic[T]):
    root_src: Path
    root_dest: Path
    glob: str
    parse: Optional[Callable[[Path], T]]
    show_progress: bool

    sources: Tuple[Path, ...] = tuple()
    operations: Dict[str, Callable[[Path, T], Path]] = dict()

    def __init__(
        self,
        root_src: Union[str, os.PathLike[str]],
        glob: str,
        root_dest: Optional[Union[str, os.PathLike[str]]] = None,
        parse: Optional[Callable[[Path], T]] = None,
        show_progress: bool = False,
    ):
        self.root_src = Path(root_src)
        self.root_dest = Path(root_dest or root_src)
        self.glob = glob
        self.parse = parse
        self.show_progress = show_progress

        self.refresh()

    def refresh(self) -> Tree[T]:
        self.sources = tuple(self.root_src.glob(self.glob))

        return self

    def organize(self, *ordering: Callable[[T], Stringable]) -> Tree:
        self.operations["organize"] = lambda path, item: self.root_dest / build_parent(item, *ordering) / path.name
        self.operations.pop("flatten", None)

        return self

    def flatten(self) -> Tree:
        self.operations["flatten"] = lambda path, _: self.root_dest / path.name
        self.operations.pop("organize", None)

        return self

    def rename(self, create_filename: Callable[[T], str]) -> Tree:
        self.operations["rename"] = lambda path, item: path.parent / create_filename(item)

        return self

    def resolve(self) -> Tree[T]:
        if len(self.operations) == 0:
            return self

        # Parse files
        parsed_sources: Iterable[Tuple]
        if self.parse is not None:
            with ProcessPoolExecutor() as executor:
                parsed_sources = zip(self.sources, executor.map(self.parse, self.sources))
        else:
            parsed_sources = zip(self.sources, self.sources)

        # Perform operations
        if self.show_progress:
            parsed_sources = tqdm(parsed_sources, desc="Process files", total=len(self.sources))

        destinations = list()
        for source, parsed in parsed_sources:
            destination = source
            try:
                for operate in self.operations.values():
                    destination = operate(destination, parsed)
            except Exception as e:
                destination = self.root_dest / e.__class__.__name__ / destination.name
            destinations.append(destination)
        self.operations.clear()

        # Move files
        paths = zip(self.sources, destinations)

        if self.show_progress:
            paths = tqdm(paths, desc="Move files", total=len(self.sources))

        for source, destination in paths:
            safe_move(source, destination)
            if is_empty(source.parent, ignore_dirs=True):
                shutil.rmtree(source.parent)

        self.refresh()

        return self

    def __enter__(self) -> Tree[T]:
        return self

    def __exit__(self, *args) -> None:
        if None in args:
            self.resolve()



import os
import shutil
from pathlib import Path


def safe_move(source: Path, destination: Path) -> None:
    if source == destination:
        return

    if not destination.parent.exists():
        os.makedirs(destination.parent)
        shutil.move(str(source), str(destination))
        return

    if not destination.exists():
        shutil.move(str(source), str(destination))
        return

    stem, suffix = destination.stem, destination.suffix
    destination = destination.parent / f"{stem} (1){suffix}"

    i = 1
    while destination.exists():
        destination = destination.parent / f"{stem} ({i}){suffix}"
        i += 1

    shutil.move(str(source), str(destination))


def is_empty(folder: Path, ignore_dirs: bool = False):
    for f in folder.rglob("*"):
        if not ignore_dirs or (ignore_dirs and not f.is_dir()):
            return False
    return True



import argparse


def main() -> int:
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("pos", type=str, help="The positional arg")
    parser.add_argument("-f", "--flag1", type=int, default=0, help="This is the first flag")
    parser.add_argument("-p", "--progress", type=bool, default=True, help="Display progress")
    parser.add_argument("-v", "--verbose", type=bool, default=False, help="Display error statements")

    args = parser.parse_args()

    try:
        print(args.pos)
        print(f"The first flag is {args.flag1}")
        if args.verbose:
            print("Some more info")
    except KeyboardInterrupt:
        print("Terminated.")
        return 1

    return 0


if __name__ == "__main__":
    exit(main())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gibbon']

package_data = \
{'': ['*']}

install_requires = \
['tqdm>=4.59.0,<5.0.0']

entry_points = \
{'console_scripts': ['gibbon = gibbon.__main__:main']}

setup_kwargs = {
    'name': 'gibbon',
    'version': '0.1.0',
    'description': '',
    'long_description': '# `gibbon`\n\n[![pypi version](https://img.shields.io/pypi/v/gibbon.svg?style=flat)](https://pypi.org/pypi/gibbon/)\n[![downloads](https://pepy.tech/badge/gibbon)](https://pepy.tech/project/gibbon)\n[![build status](https://github.com/dawsonbooth/gibbon/workflows/build/badge.svg)](https://github.com/dawsonbooth/gibbon/actions?workflow=build)\n[![python versions](https://img.shields.io/pypi/pyversions/gibbon.svg?style=flat)](https://pypi.org/pypi/gibbon/)\n[![format](https://img.shields.io/pypi/format/gibbon.svg?style=flat)](https://pypi.org/pypi/gibbon/)\n[![license](https://img.shields.io/pypi/l/gibbon.svg?style=flat)](https://github.com/dawsonbooth/gibbon/blob/master/LICENSE)\n\n## Description\n\nOrganize the files in your filesystem according to their attributes.\n\n## Installation\n\nWith [Python](https://www.python.org/downloads/) installed, simply run the following command to add the package to your project.\n\n```bash\npython -m pip install gibbon\n```\n\n## Usage\n\nThe following is an example usage of the package:\n\n```python\nfrom gibbon import Tree\n\nwith Tree("/path/to/root/folder/", glob="**/*.txt") as tree:\n    tree.flatten()\n```\n\nFeel free to [check out the docs](https://dawsonbooth.github.io/gibbon/) for more information.\n\n## License\n\nThis software is released under the terms of [MIT license](LICENSE).\n',
    'author': 'Dawson Booth',
    'author_email': 'pypi@dawsonbooth.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dawsonbooth/gibbon',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
